#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <dlfcn.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <utime.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

time_t fake_offset = 0;
void *libc6 = NULL;
time_t (*real_time)(time_t *) = NULL;
int (*real_gettimeofday)(struct timeval *, struct timezone *) = NULL;
int (*real_utime)(const char *, struct utimbuf *) = NULL;
ssize_t (*real_write)(int, const void *, size_t) = NULL;
//ssize_t (*real_read)(int, void *, size_t) = NULL;
// read is slow so skip it for now, but it should change atime
// I have to check whether open() and close() change atime...
int (*real_mknod)(const char *, mode_t, dev_t) = NULL;
int (*real_truncate)(const char *, off_t) = NULL;
int (*real_ftruncate)(int, off_t) = NULL;

void _init()
{
//  printf("init\n");
  libc6 = dlopen("libc.so.6", RTLD_LAZY | RTLD_GLOBAL);
  if(!libc6)
  {
    printf("Aieee\n");
    exit(1);
  }
  real_time = (time_t (*)(time_t *))dlsym(libc6, "time");
  real_gettimeofday = (int (*)(struct timeval *, struct timezone *))dlsym(libc6, "gettimeofday");
  real_utime = (int (*)(const char *, struct utimbuf *))dlsym(libc6, "utime");
  real_write = (ssize_t (*)(int, const void *, size_t))dlsym(libc6, "write");
//  real_read = (ssize_t (*)(int, void *, size_t))dlsym(libc6, "read");
  real_mknod = (int (*)(const char *, mode_t, dev_t))dlsym(libc6, "mknod");
  real_truncate = (int (*)(const char *, off_t))dlsym(libc6, "truncate");
  real_ftruncate = (int (*)(int, off_t))dlsym(libc6, "ftruncate");

  char *offset_buf = getenv("FAKEOFFSET");
  if(offset_buf)
  {
    fake_offset = atoi(offset_buf);
    return;
  }
  FILE *fp = fopen("/etc/faketime.conf", "r");
  if(!fp)
    return;
  char id[7];
  sprintf(id, "%d:", geteuid());
  char buf[1024];
  while(fgets(buf, sizeof(buf), fp))
  {
    if(strncmp(buf, id, strlen(id)) == 0)
    {
      fclose(fp);
      fake_offset = atoi(buf + strlen(id));
      return;
    }
  }
  fclose(fp);
}

void _fini()
{
  dlclose(libc6);
  libc6 = NULL;
  real_time = NULL;
}

const char *file_name_proc(int fd)
{
  static char buf[15];
  sprintf(buf, "/proc/%d/fd/%d", (int)getpid(), fd);
  return buf;
}

int truncate(const char *path, off_t length)
{
  if(truncate(path, length) == -1)
    return -1;
  utime(path, NULL);
  return 0;
}

int ftruncate(int fd, off_t length)
{
  if(ftruncate(fd, length) == -1)
    return -1;
  utime(file_name_proc(fd), NULL);
  return 0;
}

/*int mknod(const char *pathname, mode_t mode, dev_t dev)
{
  if(real_mknod(pathname, mode, dev) == -1)
    return -1;
  utime(pathname, NULL);
  return 0;
}*/

/*ssize_t read(int fd, void *buf, size_t count)
{
  ssize_t rc = real_read(fd, buf, count);
  if(rc != -1)
  {
    struct stat sb;
    if(fstat(fd, &sb) != -1)
    {
      struct utimbuf utb;
      utb.actime = time(NULL);
      utb.modtime = sb.st_mtime;
      real_utime(file_name_proc(fd), &utb);
    }
  }
  return rc;
}*/

ssize_t write(int fd, const void *buf, size_t count)
{
  ssize_t rc = real_write(fd, buf, count);
  if(rc != -1)
  {
    utime(file_name_proc(fd), NULL);
  }
  return rc;
}

int utime(const char *filename, struct utimbuf *buf)
{
  struct utimbuf utb;
  if(!buf)
  {
    buf = &utb;
    utb.actime = utb.modtime = time(NULL);
  }
  return real_utime(filename, buf);
}

time_t time(time_t *t)
{
  if(!libc6)
    _init();
  time_t fake = real_time(NULL) + fake_offset;
  if(t)
    *t = fake;
  return fake;
}

int gettimeofday(struct timeval *tv, struct timezone *tz)
{
  if(!libc6)
    _init();
  int rc = real_gettimeofday(tv, tz);
  if(rc == -1)
    return rc;
  if(tv)
    tv->tv_sec += fake_offset;
}
